@@ -40,6 +40,66 @@ class InvalidValueError(Exception):
     pass
 
 
+class SSHClient(paramiko.SSHClient):
+
+    def handler(self, title, instructions, prompt_list):
+        answers = []
+        for prompt_, _ in prompt_list:
+            prompt = prompt_.strip().lower()
+            if prompt.startswith('password'):
+                answers.append(self.password)
+            elif prompt.startswith('verification'):
+                answers.append(self.totp)
+            else:
+                raise ValueError('Unknown prompt: {}'.format(prompt_))
+        return answers
+
+    def auth_interactive(self, username, handler):
+        if not self.totp:
+            raise ValueError('Need a verification code for 2fa.')
+        self._transport.auth_interactive(username, handler)
+
+    def _auth(self, username, password, pkey, *args):
+        self.password = password
+        saved_exception = None
+        two_factor = False
+        allowed_types = set()
+        two_factor_types = {'keyboard-interactive', 'password'}
+
+        if pkey is not None:
+            logging.info('Trying publickey authentication')
+            try:
+                allowed_types = set(
+                    self._transport.auth_publickey(username, pkey)
+                )
+                two_factor = allowed_types & two_factor_types
+                if not two_factor:
+                    return
+            except paramiko.SSHException as e:
+                saved_exception = e
+
+        if two_factor:
+            logging.info('Trying publickey 2fa')
+            return self.auth_interactive(username, self.handler)
+
+        if password is not None:
+            logging.info('Trying password authentication')
+            try:
+                self._transport.auth_password(username, password)
+                return
+            except paramiko.SSHException as e:
+                saved_exception = e
+                allowed_types = set(getattr(e, 'allowed_types', []))
+                two_factor = allowed_types & two_factor_types
+
+        if two_factor:
+            logging.info('Trying password 2fa')
+            return self.auth_interactive(username, self.handler)
+
+        assert saved_exception is not None
+        raise saved_exception
+
+
 class PrivateKey(object):
 
     max_length = 16384  # rough number
@@ -52,31 +112,46 @@ class PrivateKey(object):
     }
 
     def __init__(self, privatekey, password=None, filename=''):
-        self.privatekey = privatekey.strip()
+        self.privatekey = privatekey
         self.filename = filename
         self.password = password
         self.check_length()
+        self.iostr = io.StringIO(privatekey)
 
     def check_length(self):
         if len(self.privatekey) > self.max_length:
             raise InvalidValueError('Invalid key length.')
 
-    def get_name(self):
-        lst = self.privatekey.split(' ', 2)
-        if len(lst) > 1:
-            return self.tag_to_name.get(lst[1])
+    def parse_name(self, iostr, tag_to_name):
+        name = None
+        for line_ in iostr:
+            line = line_.strip()
+            if line and line.startswith('-----BEGIN ') and \
+                    line.endswith(' PRIVATE KEY-----'):
+                lst = line.split(' ')
+                if len(lst) == 4:
+                    tag = lst[1]
+                    if tag:
+                        name = tag_to_name.get(tag)
+                        if name:
+                            break
+        return name, len(line_)
 
     def get_pkey_obj(self):
-        name = self.get_name()
+        name, length = self.parse_name(self.iostr, self.tag_to_name)
         if not name:
             raise InvalidValueError('Invalid key {}.'.format(self.filename))
 
+        offset = self.iostr.tell() - length
+        self.iostr.seek(offset)
+        logging.debug('Reset offset to {}.'.format(offset))
+
         logging.info('Parsing {} key'.format(name))
         pkeycls = getattr(paramiko, name+'Key')
         password = to_bytes(self.password) if self.password else None
+
         try:
-            return pkeycls.from_private_key(io.StringIO(self.privatekey),
-                                            password=password)
+            return pkeycls.from_private_key(self.iostr, password=password)
         except paramiko.PasswordRequiredException:
             raise InvalidValueError('Need a password to decrypt the key.')
         except paramiko.SSHException as exc:
@@ -240,7 +315,7 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
             super(IndexHandler, self).write_error(status_code, **kwargs)
 
     def get_ssh_client(self):
-        ssh = paramiko.SSHClient()
+        ssh = SSHClient()
         ssh._system_host_keys = self.host_keys_settings['system_host_keys']
         ssh._host_keys = self.host_keys_settings['host_keys']
         ssh._host_keys_filename = self.host_keys_settings['host_keys_filename']
@@ -291,18 +366,24 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
     def get_args(self):
         hostname = self.get_hostname()
         port = self.get_port()
-        if isinstance(self.policy, paramiko.RejectPolicy):
-            self.lookup_hostname(hostname, port)
         username = self.get_value('username')
         password = self.get_argument('password', u'')
         privatekey, filename = self.get_privatekey()
+        passphrase = self.get_argument('passphrase', u'')
+        totp = self.get_argument('totp', u'')
+
+        if isinstance(self.policy, paramiko.RejectPolicy):
+            self.lookup_hostname(hostname, port)
+
         if privatekey:
-            pkey = PrivateKey(privatekey, password, filename).get_pkey_obj()
-            password = None
+            pkey = PrivateKey(privatekey, passphrase, filename).get_pkey_obj()
         else:
             pkey = None
+
+        self.ssh_client.totp = totp
         args = (hostname, port, username, password, pkey)
         logging.debug(args)
+
         return args
 
     def get_default_encoding(self, ssh):
@@ -333,7 +414,7 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
 
         chan = ssh.invoke_shell(term='xterm')
         chan.setblocking(0)
-        worker = Worker(self.loop, ssh, chan, dst_addr, self.src_addr)
+        worker = Worker(self.loop, ssh, chan, dst_addr)
         worker.encoding = self.get_default_encoding(ssh)
         return worker
 
@@ -363,8 +444,9 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
             # for testing purpose only
             raise ValueError('Uncaught exception')
 
-        self.src_addr = self.get_client_addr()
-        if len(clients.get(self.src_addr[0], {})) >= options.maxconn:
+        ip, port = self.get_client_addr()
+        workers = clients.get(ip, {})
+        if workers and len(workers) >= options.maxconn:
             raise tornado.web.HTTPError(403, 'Too many live connections.')
 
         self.check_origin()
@@ -382,7 +464,9 @@ class IndexHandler(MixinHandler, tornado.web.RequestHandler):
             logging.error(traceback.format_exc())
             self.result.update(status=str(exc))
         else:
-            workers = clients.setdefault(worker.src_addr[0], {})
+            if not workers:
+                clients[ip] = workers
+            worker.src_addr = (ip, port)
             workers[worker.id] = worker
             self.loop.call_later(DELAY, recycle_worker, worker)
             self.result.update(id=worker.id, encoding=worker.encoding)
@@ -413,10 +497,7 @@ class WsockHandler(MixinHandler, tornado.websocket.WebSocketHandler):
             worker = workers.get(worker_id)
             if worker:
                 workers[worker_id] = None
-                try:
-                    self.set_nodelay(True)
-                except AssertionError:  # tornado 6.00 bug
-                    pass
+                self.set_nodelay(True)
                 worker.set_handler(self)
                 self.worker_ref = weakref.ref(worker)
                 self.loop.add_handler(worker.fd, worker, IOLoop.READ)
@@ -447,15 +528,9 @@ class WsockHandler(MixinHandler, tornado.websocket.WebSocketHandler):
             worker.on_write()
 
     def on_close(self):
-        if self.close_reason:
-            logging.info(
-                'Disconnecting to {}:{} with reason: {reason}'.format(
-                    *self.src_addr, reason=self.close_reason
-                )
-            )
-        else:
+        logging.info('Disconnected from {}:{}'.format(*self.src_addr))
+        if not self.close_reason:
             self.close_reason = 'client disconnected'
-            logging.info('Disconnected from {}:{}'.format(*self.src_addr))
 
         worker = self.worker_ref() if self.worker_ref else None
         if worker:
